// project created on 30/11/2002 at 22:00
//
// Author:
//     Francisco Figueiredo Jr. <fxjrlists@yahoo.com>
//
//    Copyright (C) 2002 The Npgsql Development Team
//    npgsql-general@gborg.postgresql.org
//    http://gborg.postgresql.org/project/npgsql/projdisplay.php
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using Npgsql;
using System.Data;
using System.Resources;
using NUnit.Framework;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Text.RegularExpressions;
using NpgsqlTypes;

namespace Npgsql.Tests
{
    [TestFixture]
    public class ConnectionTests : TestBase
    {
        public ConnectionTests(string backendVersion) : base(backendVersion) { }

        [Test, Description("Makes sure the connection goes through the proper state lifecycle")]
        //[Timeout(5000)]
        public void BasicLifecycle()
        {
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                bool eventOpen = false, eventClosed = false, eventBroken = false;
                conn.StateChange += (s, e) =>
                {
                    if (e.OriginalState == ConnectionState.Closed && e.CurrentState == ConnectionState.Open)
                        eventOpen = true;
                    if (e.OriginalState == ConnectionState.Open && e.CurrentState == ConnectionState.Closed)
                        eventClosed = true;
                };

                Assert.That(conn.State, Is.EqualTo(ConnectionState.Closed));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));

                // TODO: Connecting state?

                conn.Open();
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.Connector.State, Is.EqualTo(ConnectorState.Ready));
                Assert.That(eventOpen, Is.True);

                using (var cmd = new NpgsqlCommand("SELECT 1", conn))
                using (var reader = cmd.ExecuteReader())
                {
                    reader.Read();
                    Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open | ConnectionState.Fetching));
                    Assert.That(conn.State, Is.EqualTo(ConnectionState.Open));
                    Assert.That(conn.Connector.State, Is.EqualTo(ConnectorState.Fetching));
                }

                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.Connector.State, Is.EqualTo(ConnectorState.Ready));

                using (var cmd = CreateSleepCommand(conn, 1))
                {
                    var exitFlag = false;
                    var pollingTask = Task.Factory.StartNew(() =>
                    {
                        while (true)
                        {
                            if (exitFlag) {
                                Assert.Fail("Connection did not reach the Executing state");
                            }
                            if (conn.Connector.State == ConnectorState.Executing)
                            {
                                Assert.That(conn.FullState & ConnectionState.Executing, Is.Not.EqualTo(0));
                                Assert.That(conn.State, Is.EqualTo(ConnectionState.Open));
                                return;
                            }
                        }
                    });
                    cmd.ExecuteNonQuery();
                    exitFlag = true;
                    pollingTask.Wait();
                }

                conn.Close();
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Closed));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));
                Assert.That(eventClosed, Is.True);

                conn.Open();
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open));
                Assert.That(conn.Connector.State, Is.EqualTo(ConnectorState.Ready));

                // Use another connection to kill our connection
                ExecuteNonQuery(string.Format("SELECT pg_terminate_backend({0})", conn.ProcessID));

                conn.StateChange += (sender, args) =>
                {
                    if (args.CurrentState == ConnectionState.Closed)
                        eventBroken = true;
                };
                Assert.That(() => ExecuteScalar("SELECT 1", conn), Throws.Exception.TypeOf<IOException>());
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Closed));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Broken));
                Assert.That(eventBroken, Is.True);
            }
        }

        #region Connection Errors

        [Test]
        [TestCase(true,  TestName = "Pooled")]
        [TestCase(false, TestName = "NonPooled")]
        public void ConnectionRefused(bool pooled)
        {
            using (var conn = new NpgsqlConnection("Server=127.0.0.1;Port=44444;Database=d;User Id=x;Password=y" + (pooled ? "" : ";Pooling=false"))) {
                Assert.That(() => conn.Open(), Throws.Exception.TypeOf<SocketException>());
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));
            }
        }

        [Test]
        [Ignore("Fails in a non-determinstic manner and only on the build server... investigate...")]
        public void InvalidUserId()
        {
            using (var conn = new NpgsqlConnection(ConnectionString + ";userid=npgsql_tes;pooling=false"))
            {
                Assert.That(conn.Open, Throws.Exception
                    .TypeOf<NpgsqlException>()
                    .With.Property("Code").EqualTo("28P01")
                );
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));
            }
        }

        [Test, Description("Connects with a bad password to ensure the proper error is thrown")]
        public void AuthenticationFailure()
        {
            var badConnString = Regex.Replace(ConnectionString, @"Password=\w+", "Password=bad_password");
            using (var conn = new NpgsqlConnection(badConnString))
            {
                Assert.That(() => conn.Open(), Throws.Exception
                    .TypeOf<NpgsqlException>()
                    .With.Property("Code").EqualTo("28P01")
                );
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));
            }
        }

        [Test, Description("Tests that mandatory connection string parameters are indeed mandatory")]
        public void MandatoryConnectionStringParams()
        {
            Assert.That(() => new NpgsqlConnection("User ID=npgsql_tests;Password=npgsql_tests;Database=npgsql_tests").Open(), Throws.Exception.TypeOf<ArgumentException>());
        }


        [Test, Description("Reuses the same connection instance for a failed connection, then a successful one")]
        public void FailConnectThenSucceed()
        {
            ExecuteNonQuery("DROP DATABASE IF EXISTS foo");
            try
            {
                var csb = new NpgsqlConnectionStringBuilder(ConnectionString) {
                    Database = "foo",
                    Pooling = false
                };

                using (var conn = new NpgsqlConnection(csb))
                {
                    Assert.That(() => conn.Open(),
                        Throws.Exception.TypeOf<NpgsqlException>()
                        .With.Property("Code").EqualTo("3D000") // database doesn't exist
                    );
                    Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Closed));

                    // Create the database with the other connection
                    ExecuteNonQuery("CREATE DATABASE foo TEMPLATE template0");

                    conn.Open();
                    conn.Close();
                }
            }
            finally
            {
                ExecuteNonQuery("DROP DATABASE IF EXISTS foo");
            }
        }

        #endregion

        #region Notification

        [Test, Description("Simple synchronous LISTEN/NOTIFY scenario")]
        public void NotificationSync()
        {
            var receivedNotification = false;
            ExecuteNonQuery("LISTEN notifytest");
            Conn.Notification += (o, e) => receivedNotification = true;
            ExecuteNonQuery("NOTIFY notifytest");
            Assert.IsTrue(receivedNotification);
        }

        [Test, Description("An asynchronous LISTEN/NOTIFY scenario")]
        [Timeout(10000)]
        public void NotificationAsync()
        {
            var mre = new ManualResetEvent(false);
            using (var listeningConn = new NpgsqlConnection(ConnectionString + ";ContinuousProcessing=true"))
            {
                listeningConn.Open();
                ExecuteNonQuery("LISTEN notifytest2", listeningConn);
                listeningConn.Notification += (o, e) => mre.Set();

                // Send notify via the other connection
                ExecuteNonQuery("NOTIFY notifytest2");
                mre.WaitOne();

                // And again
                mre.Reset();
                ExecuteNonQuery("NOTIFY notifytest2");
                mre.WaitOne();
            }
        }

        [Test, Description("A notification arriving while we have an open Reader")]
        public void NotificationDuringReader()
        {
            var receivedNotification = false;
            using (var listeningConn = new NpgsqlConnection(ConnectionString + ";ContinuousProcessing=true"))
            {
                listeningConn.Open();
                ExecuteNonQuery("LISTEN notifytest2", listeningConn);
                listeningConn.Notification += (o, e) => receivedNotification = true;

                using (var cmd = new NpgsqlCommand("SELECT 1", listeningConn))
                using (cmd.ExecuteReader()) {
                    // Send notify via the other connection
                    ExecuteNonQuery("NOTIFY notifytest2");
                    Thread.Sleep(500);
                }
            }
            Assert.That(receivedNotification, Is.True);
        }

        [Test, Description("Receive an asynchronous notification when a message has already been prepended")]
        [Timeout(10000)]
        public void NotificationAsyncWithPrepend()
        {
            var mre = new ManualResetEvent(false);
            using (var listeningConn = new NpgsqlConnection(ConnectionString + ";ContinuousProcessing=true"))
            {
                listeningConn.Open();
                ExecuteNonQuery("LISTEN notifytest2", listeningConn);
                listeningConn.BeginTransaction();

                // Send notify via the other connection
                listeningConn.Notification += (o, e) => mre.Set();
                ExecuteNonQuery("NOTIFY notifytest2");
                mre.WaitOne();
            }
        }

        [Test, Description("Generates a notification that arrives after reader data that is already being read")]
        [IssueLink("https://github.com/npgsql/npgsql/issues/252")]
        public void NotificationAfterData()
        {
            var receivedNotification = false;
            using (var cmd = Conn.CreateCommand())
            {
                cmd.CommandText = "LISTEN notifytest1";
                cmd.ExecuteNonQuery();
                Conn.Notification += (o, e) => receivedNotification = true;

                cmd.CommandText = "SELECT generate_series(1,10000)";
                using (var reader = cmd.ExecuteReader()) {

                    //After "notify notifytest1", a notification message will be sent to client,
                    //And so the notification message will stick with the last response message of "select generate_series(1,10000)" in Npgsql's tcp receiving buffer.
                    using (var connection = new NpgsqlConnection(ConnectionString)) {
                        connection.Open();
                        using (var command = connection.CreateCommand()) {
                            command.CommandText = "NOTIFY notifytest1";
                            command.ExecuteNonQuery();
                        }
                    }

                    Assert.IsTrue(reader.Read());
                    Assert.AreEqual(1, reader.GetValue(0));
                }

                Assert.That(ExecuteScalar("SELECT 1"), Is.EqualTo(1));
                Assert.IsTrue(receivedNotification);
            }
        }

        #endregion

        #region Keepalive

        [Test, Description("Makes sure that if keepalive is enabled, broken connections are detected")]
        [Timeout(10000)]
        public void Keepalive()
        {
            var mre = new ManualResetEvent(false);
            using (var conn = new NpgsqlConnection(ConnectionString + ";KeepAlive=1;ContinuousProcessing=true"))
            {
                conn.Open();

                conn.StateChange += (sender, args) =>
                {
                    if (args.CurrentState == ConnectionState.Closed)
                        mre.Set();
                };

                // Use another connection to kill our keepalive connection
                ExecuteNonQuery(string.Format("SELECT pg_terminate_backend({0})", conn.ProcessID));
                mre.WaitOne();
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Closed));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Broken));
            }
        }

        #endregion

        [Test, IssueLink("https://github.com/npgsql/npgsql/issues/903")]
        public void DataSource()
        {
            using (var conn = new NpgsqlConnection(ConnectionString))
                Assert.That(conn.DataSource, Is.EqualTo(String.Format("tcp://{0}:{1}", conn.Host, conn.Port)));
        }

        [Test]
        [IssueLink("https://github.com/npgsql/npgsql/issues/703")]
        public void NoDatabaseDefaultsToUsername()
        {
            var csb = new NpgsqlConnectionStringBuilder(ConnectionString) { Database = null };
            using (var conn = new NpgsqlConnection(csb))
            {
                conn.Open();
                Assert.That(ExecuteScalar("SELECT current_database()"), Is.EqualTo(csb.Username));
            }
        }

        [Test, Description("Breaks a connector while it's in the pool, with a keepalive and without")]
        [TestCase(false, TestName = "WithoutKeepAlive")]
        [TestCase(false, TestName = "WithKeepAlive")]
        public void BreakConnectorInPool(bool keepAlive)
        {
            using (var conn = new NpgsqlConnection(ConnectionString + ";MaxPoolSize=1" + (keepAlive ? ";KeepAlive=1" : "")))
            {
                conn.Open();
                var connectorId = conn.ProcessID;
                conn.Close();

                // Use another connection to kill the connector currently in the pool
                ExecuteNonQuery(string.Format("SELECT pg_terminate_backend({0})", connectorId));

                conn.Open();
                Assert.That(conn.ProcessID, Is.EqualTo(connectorId));
                Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open));
                if (keepAlive)
                    Assert.That(ExecuteScalar("SELECT 1", conn), Is.EqualTo(1));
                else
                    Assert.That(() => ExecuteScalar("SELECT 1", conn), Throws.Exception);
            }
        }

        [Test]
        public void ChangeDatabase()
        {
            Conn.ChangeDatabase("template1");
            var command = new NpgsqlCommand("select current_database()", Conn);
            var result = (String)command.ExecuteScalar();
            Assert.AreEqual("template1", result);
        }

        [Test]
        public void ChangeDatabaseTestConnectionCache()
        {
            using (var conn1 = new NpgsqlConnection(ConnectionString))
            using (var conn2 = new NpgsqlConnection(ConnectionString))
            {
                //    connection 1 change database
                conn1.Open();
                conn1.ChangeDatabase("template1");
                var command = new NpgsqlCommand("select current_database()", conn1);
                var db1 = (String)command.ExecuteScalar();
                Assert.AreEqual("template1", db1);

                //    connection 2 's database should not changed, so should different from conn1
                conn2.Open();
                command = new NpgsqlCommand("select current_database()", conn2);
                var db2 = (String)command.ExecuteScalar();
                Assert.AreNotEqual(db1, db2);
            }
        }

        [Test]
        public void NestedTransaction()
        {
            Conn.BeginTransaction();
            Assert.That(() => Conn.BeginTransaction(), Throws.TypeOf<NotSupportedException>());
        }

        [Test]
        public void BeginTransactionBeforeOpen()
        {
            using (var conn = new NpgsqlConnection())
            {
                Assert.That(() => conn.BeginTransaction(), Throws.Exception.TypeOf<InvalidOperationException>());
            }
        }

        [Test]
        public void SequencialTransaction()
        {
            Conn.BeginTransaction().Rollback();
            Conn.BeginTransaction();
        }

        [Test, Description("Tests closing a connector while a reader is open")]
        [TestCase(true, TestName = "Pooled")]
        [TestCase(false, TestName = "NonPooled")]
        [Timeout(10000)]
        public void CloseDuringRead(bool pooled)
        {
            var conn = new NpgsqlConnection(ConnectionString + ";" + (pooled ? "MaxPoolSize=1" : "Pooling=false"));
            conn.Open();
            var connectorId = conn.ProcessID;
            using (var cmd = new NpgsqlCommand("SELECT 1", conn))
            using (var reader = cmd.ExecuteReader())
            {
                reader.Read();
                conn.Close();
                Assert.That(conn.State, Is.EqualTo(ConnectionState.Closed));
                Assert.That(reader.IsClosed);
            }

            conn.Open();
            if (pooled)   // Make sure we can reuse the pooled connector
                Assert.That(conn.ProcessID, Is.EqualTo(connectorId));
            Assert.That(conn.FullState, Is.EqualTo(ConnectionState.Open));
            Assert.That(ExecuteScalar("SELECT 1"), Is.EqualTo(1));
        }

        [Test]
        public void SearchPathSupport()
        {
            using (var conn = new NpgsqlConnection(ConnectionString + ";searchpath=public"))
            {
                conn.Open();
                var c = new NpgsqlCommand("show search_path", conn);
                var searchpath = (String)c.ExecuteScalar();
                //Note, public is no longer implicitly added to paths, so this is no longer "public, public".
                Assert.AreEqual("public", searchpath);
            }
        }

        [Test]
        public void ConnectorNotInitializedException1000581()
        {
            var command = new NpgsqlCommand();
            command.CommandText = @"SELECT 123";

            for (var i = 0; i < 2; i++)
            {
                using (var connection = new NpgsqlConnection(ConnectionString))
                {
                    connection.Open();
                    command.Connection = connection;
                    command.Transaction = connection.BeginTransaction();
                    command.ExecuteScalar();
                    command.Transaction.Commit();
                }
            }
        }

        [Test]
        [Ignore]
        public void NpgsqlErrorRepro1()
        {
            throw new NotImplementedException();
#if WHAT_TO_DO_WITH_THIS
            using (var connection = new NpgsqlConnection(ConnectionString))
            {
                connection.Open();
                using (var transaction = connection.BeginTransaction())
                {
                    var largeObjectMgr = new LargeObjectManager(connection);
                    try
                    {
                        var largeObject = largeObjectMgr.Open(-1, LargeObjectManager.READWRITE);
                        transaction.Commit();
                    }
                    catch
                    {
                        // ignore the LO failure
                    }
                } // *1* sometimes it throws "System.NotSupportedException: This stream does not support seek operations"

                using (var command = connection.CreateCommand())
                {
                    command.CommandText = "SELECT * FROM pg_database";
                    using (var reader = command.ExecuteReader())
                    {
                        Assert.IsTrue(reader.Read()); // *2* this fails if the initial connection is used
                    }
                }
            } // *3* sometimes it throws "System.NotSupportedException: This stream does not support seek operations"
#endif
        }

        [Test]
        public void Bug1011001()
        {
            //[#1011001] Bug in NpgsqlConnectionStringBuilder affects on cache and connection pool

            var csb1 = new NpgsqlConnectionStringBuilder(@"Server=server;Port=5432;User Id=user;Password=passwor;Database=database;");
            var cs1 = csb1.ToString();
            var csb2 = new NpgsqlConnectionStringBuilder(cs1);
            var cs2 = csb2.ToString();
            Assert.IsTrue(cs1 == cs2);
        }

        [Test]
        public void NpgsqlErrorRepro2()
        {
#if WHAT_TO_DO_WITH_THIS
            var connection = new NpgsqlConnection(ConnectionString);
            connection.Open();
            var transaction = connection.BeginTransaction();
            var largeObjectMgr = new LargeObjectManager(connection);
            try
            {
                var largeObject = largeObjectMgr.Open(-1, LargeObjectManager.READWRITE);
                transaction.Commit();
            }
            catch
            {
                // ignore the LO failure
                try
                {
                    transaction.Dispose();
                }
                catch
                {
                    // ignore dispose failure
                }
                try
                {
                    connection.Dispose();
                }
                catch
                {
                    // ignore dispose failure
                }
            }

            using (connection = new NpgsqlConnection(ConnectionString))
            {
                connection.Open();
                using (var command = connection.CreateCommand())
                {
                    command.CommandText = "SELECT * FROM pg_database";
                    using (var reader = command.ExecuteReader())
                    {
                        Assert.IsTrue(reader.Read());
                        // *1* this fails if the connection for the pool happens to be the bad one from above
                        Assert.IsTrue(!String.IsNullOrEmpty((string)reader["datname"]));
                    }
                }
            }
#endif
        }

        [Test]
        public void GetSchemaForeignKeys()
        {
            var dt = Conn.GetSchema("ForeignKeys");
            Assert.IsNotNull(dt);
        }

        [Test]
        public void GetSchemaParameterMarkerFormats()
        {
            ExecuteNonQuery("DROP TABLE IF EXISTS data; CREATE TABLE data (int INTEGER);");
            ExecuteNonQuery("INSERT INTO data (int) VALUES (4)");
            var dt = Conn.GetSchema("DataSourceInformation");
            var parameterMarkerFormat = (string)dt.Rows[0]["ParameterMarkerFormat"];

            using (var connection = new NpgsqlConnection(ConnectionString))
            {
                connection.Open();
                using (var command = connection.CreateCommand())
                {
                    const String parameterName = "@p_int";
                    command.CommandText = "SELECT * FROM data WHERE int=" + String.Format(parameterMarkerFormat, parameterName);
                    command.Parameters.Add(new NpgsqlParameter(parameterName, 4));
                    using (var reader = command.ExecuteReader())
                    {
                        Assert.IsTrue(reader.Read());
                        // This is OK, when no exceptions are occurred.
                    }
                }
            }
        }

        [Test]
        public void GetConnectionState()
        {
            // Test created to PR #164

            NpgsqlConnection c = new NpgsqlConnection();
            c.Dispose();
            Assert.AreEqual(ConnectionState.Closed, c.State);
        }

        [Test]
        public void ChangeApplicationNameWithConnectionStringBuilder()
        {
            // Test for issue #165 on github.
            NpgsqlConnectionStringBuilder builder = new NpgsqlConnectionStringBuilder();
            builder.ApplicationName = "test";
        }

        [Test, Description("Makes sure notices are probably received and emitted as events")]
        public void Notice()
        {
            // Make sure messages are in English
            ExecuteNonQuery(@"SET lc_messages='English_United States.1252'");
            ExecuteNonQuery(@"
                 CREATE OR REPLACE FUNCTION emit_notice() RETURNS VOID AS
                    'BEGIN RAISE NOTICE ''testnotice''; END;'
                 LANGUAGE 'plpgsql';
            ");

            NpgsqlNotice notice = null;
            NoticeEventHandler action = (sender, args) => notice = args.Notice;
            Conn.Notice += action;
            try
            {
                ExecuteNonQuery("SELECT emit_notice()::TEXT");  // See docs for CreateSleepCommand
                Assert.That(notice, Is.Not.Null, "No notice was emitted");
                Assert.That(notice.MessageText, Is.EqualTo("testnotice"));
                Assert.That(notice.Severity, Is.EqualTo("NOTICE"));
            }
            finally
            {
                Conn.Notice -= action;
            }
        }

        [Test, Description("Makes sure that concurrent use of the connection throws an exception")]
        public void ConcurrentUse()
        {
            using (var cmd = new NpgsqlCommand("SELECT 1", Conn))
            using (cmd.ExecuteReader())
                Assert.That(() => ExecuteScalar("SELECT 1", Conn), Throws.Exception.TypeOf<InvalidOperationException>());
        }

        [Test]
        public void NoContinuousProcessingWithSslStream()
        {
            using (var conn = new NpgsqlConnection(ConnectionString + ";UseSslStream=true;ContinuousProcessing=true"))
                Assert.That(() => conn.Open(), Throws.Exception.TypeOf<ArgumentException>());
        }

        [Test]
        [IssueLink("https://github.com/npgsql/npgsql/issues/743")]
        public void Clone()
        {
            var connString = new NpgsqlConnectionStringBuilder(ConnectionString) { Pooling = false };
            using (var conn = new NpgsqlConnection(connString))
            {
                ProvideClientCertificatesCallback callback1 = certificates => { };
                conn.ProvideClientCertificatesCallback = callback1;
                RemoteCertificateValidationCallback callback2 = (sender, certificate, chain, errors) => true;
                conn.UserCertificateValidationCallback = callback2;

                conn.Open();
                using (var conn2 = (NpgsqlConnection)((ICloneable)conn).Clone())
                {
                    Assert.That(conn2.ConnectionString, Is.EqualTo(conn.ConnectionString));
                    Assert.That(conn2.ProvideClientCertificatesCallback, Is.SameAs(callback1));
                    Assert.That(conn2.UserCertificateValidationCallback, Is.SameAs(callback2));
                    conn2.Open();
                }
            }
        }

        [Test, IssueLink("https://github.com/npgsql/npgsql/issues/824")]
        public void ReloadTypes()
        {
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
                Assert.That(ExecuteScalar("SELECT EXISTS (SELECT * FROM pg_type WHERE typname='reload_types_enum')"), Is.False);
                ExecuteNonQuery("CREATE TYPE pg_temp.reload_types_enum AS ENUM ('First', 'Second')");
                conn.ReloadTypes();
                conn.RegisterEnum<ReloadTypesEnum>("reload_types_enum");
            }
        }
        enum ReloadTypesEnum { First, Second };

        [Test]
        [IssueLink("https://github.com/npgsql/npgsql/issues/736")]
        public void ManyOpenClose()
        {
            // The connector's _sentRfqPrependedMessages is a byte, too many open/closes made it overflow
            for (var i = 0; i < 255; i++)
            {
                using (var conn = new NpgsqlConnection(ConnectionString))
                {
                    conn.Open();
                }
            }
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
            }
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
                Assert.That(ExecuteScalar("SELECT 1", conn), Is.EqualTo(1));
            }
        }

        [Test]
        [IssueLink("https://github.com/npgsql/npgsql/issues/736")]
        public void ManyOpenCloseWithTransaction()
        {
            // The connector's _sentRfqPrependedMessages is a byte, too many open/closes made it overflow
            for (var i = 0; i < 255; i++)
            {
                using (var conn = new NpgsqlConnection(ConnectionString))
                {
                    conn.Open();
                    conn.BeginTransaction();
                }
            }
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
                Assert.That(ExecuteScalar("SELECT 1", conn), Is.EqualTo(1));
            }
        }

        [Test]
        [IssueLink("https://github.com/npgsql/npgsql/issues/927")]
        [IssueLink("https://github.com/npgsql/npgsql/issues/736")]
        public void RollbackOnClose()
        {
            // Npgsql 3.0.0 to 3.0.4 prepended a rollback for the next time the connector is used, as an optimization.
            // This caused some issues (#927) and was removed.
            int processId;
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
                processId = conn.Connector.BackendProcessId;
                conn.BeginTransaction();
                ExecuteNonQuery("SELECT 1", conn);
                Assert.That(conn.Connector.TransactionStatus, Is.EqualTo(TransactionStatus.InTransactionBlock));
            }
            using (var conn = new NpgsqlConnection(ConnectionString))
            {
                conn.Open();
                Assert.That(conn.Connector.BackendProcessId, Is.EqualTo(processId));
                Assert.That(conn.Connector.TransactionStatus, Is.EqualTo(TransactionStatus.Idle));
            }
        }

        [Test, Description("Tests an exception happening when sending the Terminate message while closing a ready connector")]
        [IssueLink("https://github.com/npgsql/npgsql/issues/777")]
        public void ExceptionDuringClose()
        {
            var connString = new NpgsqlConnectionStringBuilder(ConnectionString) { Pooling = false };
            using (var conn = new NpgsqlConnection(connString))
            {
                conn.Open();
                var connectorId = conn.ProcessID;

                // Use another connection to kill our connector
                ExecuteNonQuery(string.Format("SELECT pg_terminate_backend({0})", connectorId));

                conn.Close();
            }
        }

        #region GetSchema

        [Test]
        public void GetSchema()
        {
            using (NpgsqlConnection c = new NpgsqlConnection())
            {
                DataTable metaDataCollections = c.GetSchema();
                Assert.IsTrue(metaDataCollections.Rows.Count > 0, "There should be one or more metadatacollections returned. No connectionstring is required.");
            }
        }

        [Test]
        public void GetSchemaWithDbMetaDataCollectionNames()
        {
            DataTable metaDataCollections = Conn.GetSchema(System.Data.Common.DbMetaDataCollectionNames.MetaDataCollections);
            Assert.IsTrue(metaDataCollections.Rows.Count > 0, "There should be one or more metadatacollections returned.");
            foreach (DataRow row in metaDataCollections.Rows)
            {
                var collectionName = (string)row["CollectionName"];
                //checking this collection
                if (collectionName != System.Data.Common.DbMetaDataCollectionNames.MetaDataCollections)
                {
                    var collection = Conn.GetSchema(collectionName);
                    Assert.IsNotNull(collection, "Each of the advertised metadata collections should work");
                }
            }
        }

        [Test]
        public void GetSchemaWithRestrictions()
        {
            DataTable metaDataCollections = Conn.GetSchema(System.Data.Common.DbMetaDataCollectionNames.Restrictions);
            Assert.IsTrue(metaDataCollections.Rows.Count > 0, "There should be one or more Restrictions returned.");
        }

        [Test]
        public void GetSchemaWithReservedWords()
        {
            DataTable metaDataCollections = Conn.GetSchema(System.Data.Common.DbMetaDataCollectionNames.ReservedWords);
            Assert.IsTrue(metaDataCollections.Rows.Count > 0, "There should be one or more ReservedWords returned.");
        }

        #endregion
    }
}
